#ifndef AWS_IO_URI_H
#define AWS_IO_URI_H
/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 */
#include <aws/io/io.h>

#include <aws/common/uri.h>

#endif /* AWS_IO_URI_H */
